from datetime import datetime as dt

class Timer:
    def __init__(self):
        self.factor=1
        self.startTime=0.0

    def now(self):
        return dt.now().timestamp()

    def start(self, factor=1, offset=0):
        self.factor=factor
        self.startTime=self.now()+offset

    def get(self):
        if self.startTime>0:
            return (self.now()-self.startTime)*self.factor
        else:
            return 0

    def clear(self):
        t=self.get()
        self.startTime=0
        return t





try:
    import sounddevice as sd
    import numpy as np
except:
    pass
from pydub import AudioSegment

from .driver_audio import driver_audio

class driver_sounddevice_audio(driver_audio):
    def __init__(self):
        super().__init__()

    def play(self, start=0, end=0):
        super().play(start=start, end=end)
        buffer=self.audio.get_array_of_samples()
        if end==0:
            return sd.play(buffer[int(start):], self.audio.frame_rate)
        return sd.play(buffer[int(start):int(end)], self.audio.frame_rate)

    def stop(self):
        sd.stop()
        super().stop()

    def rec(self):
        super().rec()
        len=self.fps*self.channels*60*10
        self.record_buffer=sd.rec(len, self.fps, channels=self.channels)

    def wait(self):
        super().wait()
        return sd.wait()

    def setAudio(self, buffer, fps, sample_width, channels):
        if not isinstance(buffer, np.ndarray):
            return
            raise ValueError("audio_data must be a NumPy array")
        if buffer.dtype not in [np.float32, np.int16, np.float64]:
            return
            raise ValueError(f"audio data must be float64, float32 or int16 ({buffer.dtype})")
        buf=buffer
        # Normalize audio data to appropriate range
        if buffer.dtype in [ np.float32, np.float64 ]:
            buf = np.clip(buffer, -1.0, 1.0) * np.iinfo(np.int16).max
        buf = buf.astype(np.int16)
        # Create a pydub AudioSegment from the NumPy array
        return AudioSegment(buf.tobytes(),
            frame_rate=fps, sample_width=sample_width, channels=channels)





#!/usr/bin/python3
import sys, os, fcntl, select, asyncio, termios, tty, logging, pyte, re, icat

"""
         0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
U+250x   ─   ━   │   ┃   ┄   ┅   ┆   ┇   ┈   ┉   ┊   ┋   ┌   ┍   ┎   ┏
U+251x   ┐   ┑   ┒   ┓   └   ┕   ┖   ┗   ┘   ┙   ┚   ┛   ├   ┝   ┞   ┟
U+252x   ┠   ┡   ┢   ┣   ┤   ┥   ┦   ┧   ┨   ┩   ┪   ┫   ┬   ┭   ┮   ┯
U+253x   ┰   ┱   ┲   ┳   ┴   ┵   ┶   ┷   ┸   ┹   ┺   ┻   ┼   ┽   ┾   ┿
U+254x   ╀   ╁   ╂   ╃   ╄   ╅   ╆   ╇   ╈   ╉   ╊   ╋   ╌   ╍   ╎   ╏

         0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
U+255x   ═   ║   ╒   ╓   ╔   ╕   ╖   ╗   ╘   ╙   ╚   ╛   ╜   ╝   ╞   ╟
U+256x   ╠   ╡   ╢   ╣   ╤   ╥   ╦   ╧   ╨   ╩   ╪   ╫   ╬   ╭   ╮   ╯
U+257x   ╰   ╱   ╲   ╳   ╴   ╵   ╶   ╷   ╸   ╹   ╺   ╻   ╼   ╽   ╾   ╿

         0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
U+258x   ▀   ▁   ▂   ▃   ▄   ▅   ▆   ▇   █   ▉   ▊   ▋   ▌   ▍   ▎   ▏
U+259x   ▐   ░   ▒   ▓   ▔   ▕   ▖   ▗   ▘   ▙   ▚   ▛   ▜   ▝   ▞   ▟
"""

grchr={}
grchr['ascii']={'hline':'-', 'vline':'|',
                'TH':'^', 'BH':'o',
                'B0':' ', 'B25':':', 'BN60':'$', 'B75':'#', 'B100':'@',
                'BLC':'\\', 'TLC':'/', 'BRC':'/', 'TRC':'\\',
                'BLB':'+', 'TLB':'+', 'BRB':'+', 'TRB':'+',
                'TBR':'|', 'TBL':'|', 'BLR':'-', 'TLR':'-', 'TBLR':'+',
                }

grchr['utf8']={ 'hline':'\u2500', 'vline':'\u2502',
                'TH':'\u2580', 'BH':'\u2584',
                'B0':' ', 'B25':'\u2591', 'B50':'\u2593', 'B75':'\u2593', 'B100':'\u2588',
                'BLC':'\u256E', 'TLC':'\u256F', 'BRC':'\u256D', 'TRC':'\u2570',
                'BLB':'\u2510', 'TLB':'\u2518', 'BRB':'\u250C', 'TRB':'\u2514',
                'TBR':'\u251C', 'TBL':'\u2524', 'BLR':'\u252C', 'TLR':'\u2534', 'TBLR':'\u253C',
               }

theme={}
theme['inside']={
        'TL': 'BH', 'TC': 'BH', 'TR': 'BH',
        'ML': 'B100', 'MC': 'B75', 'MR': 'B100',
        'BL': 'TH', 'BC': 'TH', 'BR': 'TH'
        }

theme['outside']={
        'TL': 'B100', 'TC': 'TH', 'TR': 'B100',
        'ML': 'B100', 'MC': 'B0', 'MR': 'B100',
        'BL': 'B100', 'BC': 'BH', 'BR': 'B100'
        }

theme['curve']={
        'TL': 'BRC', 'TC': 'hline', 'TR': 'BLC',
        'ML': 'vline', 'MC': 'B0', 'MR': 'vline',
        'BL': 'TRC', 'BC': 'hline', 'BR': 'TLC'
        }

rgb_file_path = '/usr/share/X11/rgb.txt'

class termcontrol:
    def __init__(self):
        self.x11_colors = self.parse_rgb_file(rgb_file_path)
        self.image_support=[]
        self.img_cache={}
        term=os.environ.get('TERM', '')
        konsole_ver=os.environ.get('KONSOLE_VERSION', '')
        if 'kitty' in term:
            self.image_support.append('kitty')
        if 'vt340' in term or len(konsole_ver or '')>0:
            self.image_support.append('sixel')

    def enable_mouse(self, utf8=True):
        if(utf8):
            return "\x1b[?1005h"
        return "\x1b[?1000h"

    def disable_mouse(self, utf8=True):
        if(utf8):
            return "\x1b[?1005l"
        return "\x1b[?1000l"

    def enable_cursor(self):
        return "\x1b[?25h"

    def disable_cursor(self):
        return "\x1b[?25l"

    def normal_screen(self):
        return "\x1b[?1049l"

    def alt_screen(self):
        return "\x1b[?1049h"

    def set_title(self, title):
        return f"\x1b]0;{title}\\a"

    def pause_terminal_output(self):
        sys.stdout.flush()
        os.system('stty -icanon -echo')

    def resume_terminal_output(self):
        sys.stdout.flush()
        os.system('stty icanon echo')

    def parse_rgb_file(self, file_path):
        colors = {}
        if not os.path.isfile(file_path):
            return colors
        return colors
        with open(file_path, 'r') as file:
            for line in file:
                if not line.startswith('!'):
                    parts = line.strip().split('\t')
                    if len(parts) >= 4:
                        name = parts[3].lower()
                        red, green, blue = int(parts[0]), int(parts[1]), int(parts[2])
                        colors[name] = {'red':red, 'green':green, 'blue':blue}
        return colors

    def pause(self):
        print ('[pause]')
        return sys.stdin.readline()

    def get_terminal_size(self):
        import array, fcntl, sys, termios
        buf = array.array('H', [0, 0, 0, 0])
        fcntl.ioctl(sys.stdout, termios.TIOCGWINSZ, buf)
        # Create a dictionary with meaningful keys
        window_info = {
            "rows": buf[0],
            "columns": buf[1],
            "width": buf[2],
            "height": buf[3]
        }
        return window_info

    def color(self, color):
        if type(color)==int:
            return color
        co={
                'black'  : 0,
                'red'    : 1,
                'green'  : 2,
                'yellow' : 3,
                'brown'  : 3,
                'blue'   : 4,
                'magenta': 5,
                'cyan'   : 6,
                'white'  : 7,
                'brightblack'  : 8,
                'brightred'    : 9,
                'brightgreen'  : 10,
                'brightbrown'  : 11,
                'brightyellow' : 11,
                'brightblue'   : 12,
                'brightmagenta': 13,
                'brightcyan'   : 14,
                'brightwhite'  : 15,
            }
        if type(color)==str:
            regex = r'^([A-Fa-f0-9]{6})$'
            if re.match(regex, color) is not None:
                color={
                        'red'  :int(color[0:2], 16),
                        'green':int(color[2:4], 16),
                        'blue' :int(color[4:6], 16),
                      }
                return color
            regex = r'^([A-Fa-f0-9]{3})$'
            if re.match(regex, color) is not None:
                color={
                        'red'  :int(color[0:1], 16)*16,
                        'green':int(color[1:2], 16)*16,
                        'blue' :int(color[2:3], 16)*16,
                      }
                return color
            regex = r'^#([A-Fa-f0-9]{6})$'
            if re.match(regex, color) is not None:
                color={
                        'red'  :int(color[1:3], 16),
                        'green':int(color[3:5], 16),
                        'blue' :int(color[5:7], 16),
                      }
                return color
            regex = r'^#([A-Fa-f0-9]{3})$'
            if re.match(regex, color) is not None:
                color={
                        'red'  :int(color[1:2], 16)*16,
                        'green':int(color[2:3], 16)*16,
                        'blue' :int(color[3:4], 16)*16,
                      }
                return color
            if co.get(color):
                return co.get(color)
            return self.x11_colors.get(color)
        return None

    def getRGB(self, c):
        color=self.color(c)
        if type(color)==dict:
            return color
        if type(color)==int:
            pass #FIXME
        return {'red':127, 'green':127, 'blue':127}

    def ansicolor(self, fg=7, bg=None,
                  bold=False, dim=False, italic=False, underline=False,
                  strike=False, blink=False, blink2=False, reverse=False,
                  bold_is_bright=False):
        if fg=='default':
            fg=7
        if bg=='default':
            bg=None
        fg=self.color(fg)
        bg=self.color(bg)
        fgs=""
        bgs=""
        if type(fg)==int:
            if fg<16:
                if fg<8:
                    if bold_is_bright:
                        fgs=f"{fg+90}"
                    else:
                        fgs=f"{fg+30}"
                else:
                    fgs=f"{(fg-8)+90}"
            else:
                fgs=f'38;5;{fg}'
        elif type(fg)==dict:
            fgs=f'38;2;{fg["red"]};{fg["green"]};{fg["blue"]}'
        if type(bg)==int:
            if bg<16:
                if bg<8:
                    bgs=f"{bg+40}"
                else:
                    bgs=f"{(bg-8)+100}"
            else:
                bgs=f'48;5;{bg}'
        elif type(bg)==dict:
            bgs=f'48;2;{bg["red"]};{bg["green"]};{bg["blue"]}'
        bo, bl, bl2, dm, it, ul, st, rv="","","","","","","", ""
        if bold:bo='1;'
        if dim:bdm='2;'
        if italic:it='3;'
        if underline:ul='4;'
        if blink:bl='5;'
        if blink2:bl2='6;'
        if reverse:rv='7;'
        ansi=""
        if len(bgs) and len(fgs):
            ansi=f'{fgs};{bgs}'
        elif len(bgs):
            ansi=bgs
        elif len(fgs):
            ansi=fgs
        if len(ansi)>0:
            return f"\x1b[{bo}{dm}{it}{ul}{bl}{bl2}{rv}{ansi}m"
        return ""

    def drawRuler(self,w,h):
        buffer=''
        for y in range(0,h):
            for x in range(0,int((w)/10)):
                buffer+=self.gotoxy(x*10+1,y)
                buffer+=f'({x*10+1},{y})'
        return buffer

    def pyte_render(self, x, y, screen, line=1,
                    fg='default', bg='default',
                    fg0='default', bg0='default',
                    bold_is_bright=False):
        bold=False
        blink=False
        w=screen.columns
        h=screen.screen_lines
        start_line=line-1
        if start_line<0:
            start_line=int(screen.cursor.y-h+2)
        if start_line<0:
            start_line=0
        if start_line>int(screen.cursor.y-h+2):
            start_line=int(screen.cursor.y-h+2)
        buffer = self.ansicolor(fg, bg, bold=bold, blink=blink)
        screen.cursor_position(screen.screen_lines+start_line, 1)
        for yy in range(start_line, start_line+h):
            buffer += self.gotoxy(x, y+yy-(start_line))
            buffer+=self.ansicolor(fg, bg)
            for xx in range(w):
                if screen.buffer[yy][xx].fg!=fg or screen.buffer[yy][xx].bold!=bold:
                    fg=screen.buffer[yy][xx].fg
                    bold=screen.buffer[yy][xx].bold
                    buffer += self.ansicolor(fg, None, bold=bold, bold_is_bright=bold_is_bright)
                if screen.buffer[yy][xx].bg!=bg or screen.buffer[yy][xx].blink!=blink:
                    bg=screen.buffer[yy][xx].bg
                    blink=screen.buffer[yy][xx].blink
                    buffer += self.ansicolor(None, bg, blink=blink)
                buffer += screen.buffer[yy][xx].data
            buffer+=self.ansicolor(fg0, bg0)
        return buffer

    def gotoxy(self, x, y):
        return f'\x1b[{int(y)};{int(x)}f'

    def clear(self):
        return '\x1b[2J'

    def reset(self):
        return '\x1b[0m'

    def setbg(self, c):
        return self.ansicolor(None, c)

    def setfg(self, c):
        return self.ansicolor(c, None)

    def up(self, n):
        return f'\x1b[{n}A'

    def down(self, n):
        return f'\x1b[{n}B'

    def left(self, n):
        return f'\x1b[{n}D'

    def right(self, n):
        return f'\x1b[{n}C'

    def clear_images(self):
        out=''
        if 'kitty' in self.image_support:
            out+='\x1b_Ga=d\x1b\\'
        if 'sixel' in self.image_support:
            pass
        return out

    def showImage(self, image, x=0, y=0, w=30, h=15, showInfo=False, mode='auto', charset='utf8'):
        desc=""
        imgX,imgY=0,0
        if(showInfo):
            try:
                img = Image.open(image)
                imgX,imgY=img.size
                img.close()
            except:
                pass
                #logging.WARNING(f"can't open {image} as an image.")
            filename=os.path.basename(image)
            desc=f'({imgX}x{imgY}) {filename}'[:w]
            descX=int(x+(w/2)-(len(desc)/2))+1
            descY=int(y+h)-1
            desc=f'\x1b[s\x1b[48;5;245;30m\x1b[{descY};{descX}H{desc}\n'
        start_pos = f'\x1b[{y};{x+1}H'
        if not self.img_cache.get(image):
            ic=ICat(w=int(w), h=int(h), zoom='aspect', f=True, x=int(0), y=int(0), place=True, mode=mode, charset=charset)
            self.img_cache[image]=ic.print(image)
        return f'{start_pos}{self.img_cache[image]}{desc}'

def clean(input_string):
    # Use regular expressions to replace consecutive whitespace characters with a single space
    cleaned_string = re.sub(r'\s+', ' ', input_string)
    # Remove leading and trailing spaces
    return cleaned_string.strip()

class pyteLogger(logging.Logger):
    def __init__(self, refresh_class=None):
        logging.__init__('pyte')
        self.refresh_class=refresh_class

    def debug(self, msg, *args, **kwargs):
        logging.debug(clean(msg), *args, **kwargs)
        if self.refresh_class: self.refresh_class.refresh()

    def info(self, msg, *args, **kwargs):
        logging.info(clean(msg), *args, **kwargs)
        if self.refresh_class: self.refresh_class.refresh()

    def warning(self, msg, *args, **kwargs):
        logging.warning(clean(msg), *args, **kwargs)
        if self.refresh_class: self.refresh_class.refresh()

    def error(self, msg, *args, **kwargs):
        logging.error(clean(msg), *args, **kwargs)
        if self.refresh_class: self.refresh_class.refresh()

    def critical(self, msg, *args, **kwargs):
        logging.critical(clean(msg), *args, **kwargs)
        if self.refresh_class: self.refresh_class.refresh()
        exit()

class boxDraw:
    def __init__(self, bgColor=24,
                bg0=0, fg0=7,
                chars="",
                frameColors=[],
                title="", statusBar='',
                mode='auto', charset='utf8',
                style='inside',
                ):
        self.term=termcontrol()
        self.fg0, self.bg0=fg0, bg0
        self.bgColor=bgColor
        if len(chars)!=9:
            cd=grchr['utf8']
            if charset.lower() in ['utf8', 'utf-8']:
                cd=grchr['utf8']
            else:
                cd=grchr['ascii']
            self.chars=f'{cd[theme[style]["TL"]]}{cd[theme[style]["TC"]]}{cd[theme[style]["TR"]]}'\
                        f'{cd[theme[style]["ML"]]}{cd[theme[style]["MC"]]}{cd[theme[style]["MR"]]}'\
                        f'{cd[theme[style]["BL"]]}{cd[theme[style]["BC"]]}{cd[theme[style]["BR"]]}'
        else:
            self.chars=chars
        fr=False
        if len(frameColors)!=9:
            fr=True
        if mode in ['sixel', 'kitty', '24bit', '24-bit', 'auto']:
            if fr:
                self.frameColors=['#FFF', '#AAA','#777','#AAA', 0, '#555', '#777','#555','#333']
            if type(bgColor)==int and bgColor>255:
                self.bgColor=0
            else:
                self.bgColor=bgColor
        elif mode in ['8bit', '8-bit', '256color', '8bitgrey', 'grey', '8bitbright']:
            if fr:
                self.frameColors=[255, 245, 240, 245, 0, 237, 240, 237, 235]
            if type(bgColor)!=int or bgColor>255:
                self.bgColor=0
            else:
                self.bgColor=bgColor
        elif mode in ['4bit', '4-bit', '16color', '4bitgrey']:
            if fr:
                self.frameColors=[15, 7, 8, 7, 0, 8, 7, 8, 0]
            if type(bgColor)!=int or bgColor>15:
                self.bgColor=0
            else:
                self.bgColor=bgColor
        else:
            if fr:
                self.frameColors=[7, 7, 7, 7, 0, 7, 7, 7, 7]
            self.bgColor=0
        self.tinted=None
        self.title=title
        self.statusBar=statusBar

    def setColors(self, bgcolor, frameColors):
        self.bgColor=bgColor
        self.frameColors=frameColors

    def tintFrame(self, color):
        if color==None:
            self.tinted=None
            return
        c=self.term.getRGB(color)
        r, g, b=c['red'], c['green'], c['blue']
        r=r/255.0
        g=g/255.0
        b=b/255.0
        self.tinted=[]
        for i in range(0, len(self.frameColors)):
            c=self.term.getRGB(i)
            fr,fg,fb=c['red'], c['green'], c['blue']
            fr=int(fr/16*r)
            fg=int(fg/16*g)
            fb=int(fb/16*b)
            self.tinted.append(F"#{fr:X}{fg:X}{fb:X}")

    def unTintFrame(self):
        self.tinted=None

    def setCharacters(self):
        self.chars=chars

    def invert(self, cl):
        c=cl.copy()
        for i in range(0, len(cl)):
            c[i]=cl[8-i]
        return c

    def draw(self, x, y, w, h, fill=True, invert=False):
        if(w<3): w=3
        if(h<3): h=3
        colors=self.frameColors
        if(self.tinted):
            colors=self.tinted
        if invert:
            colors=self.invert(colors)
            pass
        buff=self.term.gotoxy(x,y)
        buff+=self.term.ansicolor(colors[0], self.bg0)+self.chars[0]
        buff+=self.term.ansicolor(colors[1], self.bg0)+self.chars[1]*(w-2)
        buff+=self.term.ansicolor(colors[2], self.bg0)+self.chars[2]
        buff+=self.term.ansicolor(self.fg0, self.bg0)
        for i in range(1,h-1):
            buff+=self.term.gotoxy(x,y+i)+\
                self.term.ansicolor(colors[3], self.bg0)+self.chars[3]
            if(fill):
                buff+=self.term.ansicolor(colors[4], self.bgColor)+self.chars[4]*(w-2)
            else:
                buff+=self.term.ansicolor(colors[4], self.bgColor)
                buff+=F"\x1b[{w-2}C"
            buff+=self.term.ansicolor(colors[5], self.bg0)+self.chars[5]
            buff+=self.term.ansicolor(self.fg0, self.bg0)
        buff+=self.term.gotoxy(x,y+h-1)
        buff+=self.term.ansicolor(colors[6], self.bg0)+self.chars[6]
        buff+=self.term.ansicolor(colors[7], self.bg0)+self.chars[7]*(w-2)
        buff+=self.term.ansicolor(colors[8], self.bg0)+self.chars[8]
        buff+=self.term.ansicolor(self.fg0, self.bg0)
        if self.title!='':
            desc=self.title
            descX=int(x+(w/2)-(len(desc)/2))+1
            descY=int(y)
            descPos=self.move(descX, descY)
            descColor=self.term.ansicolor(16, colors[1])
            buff+=f'{descPos}{descColor}{desc}'
            buff+=self.term.ansicolor(self.fg0, self.bg0)
        if self.statusBar!='':
            pass
        return buff

class termKeyboard:
    def __init__(self):
        self.kbtimeout=0.25
        self.keymap={"\x1b[A":"Up", "\x1b[B":"Down",\
                 "\x1b[C":"Right", "\x1b[D":"Left",\
                 "\x7f":"Backspace", "\x09":"Tab",\
                 "\x0a":"Enter", "\x1b\x1b":"Esc",\
                 "\x1b[H":"Home", "\x1b[F":"End",\
                 "\x1b[5~":"PgUp", "\x1b[6~":"PgDn",\
                 "\x1b[2~":"Ins", "\x1b[3~":"Del",\
                 "\x1bOP":"F1", "\x1bOQ":"F2",\
                 "\x1bOR":"F3", "\x1bOS":"F4",\
                 "\x1b[15~":"F5", "\x1b[17~": "F6",\
                 "\x1b[18~":"F7", "\x1b[19~": "F8",\
                 "\x1b[20~":"F9", "\x1b[21~": "F10",\
                 "\x1b[23~":"F11", "\x1b[24~": "F12",\
                 "\x1b[32~":"SyRq", "\x1b[34~": "Brk",
                 "\x1b[Z":"Shift Tab"}

    def disable_keyboard_echo(self): # Get the current terminal attributes
        attributes = termios.tcgetattr(sys.stdin)
        # Disable echo flag
        attributes[3] = attributes[3] & ~termios.ECHO
        # Apply the modified attributes
        termios.tcsetattr(sys.stdin, termios.TCSANOW, attributes)

    def enable_keyboard_echo(self): # Get the current terminal attributes
        attributes = termios.tcgetattr(sys.stdin)
        # Enable echo flag
        attributes[3] = attributes[3] | termios.ECHO
        # Apply the modified attributes
        termios.tcsetattr(sys.stdin, termios.TCSANOW, attributes)

    def read(self, bin=False):
        rlist, _, _ = select.select([sys.stdin], [], [], self.kbtimeout)
        if rlist:
            if not bin:
                try:
                    return sys.stdin.read(1)
                except:
                    return sys.stdin.buffer.read(1)
            else:
                return sys.stdin.buffer.read(1)
        return ''

    def ord(self, d):
        if(type(d)==int):
            return d
        if(type(d)==str):
            return ord(d[0])
        if(type(d)==bytes):
            return int.from_bytes(d)
        return int(d)

    def read_keyboard_input(self): # Get the current settings of the terminal
        flags = fcntl.fcntl(sys.stdin.fileno(), fcntl.F_GETFL)
        fcntl.fcntl(sys.stdin.fileno(), fcntl.F_SETFL, flags | os.O_NONBLOCK)
        filedescriptors = termios.tcgetattr(sys.stdin)
        # Set the terminal to cooked mode
        tty.setcbreak(sys.stdin)
        char = self.read()
        buffer=char
        # Check if the character is an arrow key or a function key
        if char == "\x1b":
            char = self.read()
            buffer+=char
            if(char=='O'):      #special key F1-F4
                char = self.read()
                buffer+=char
            elif char=='[':     #special key or mouse
                char = self.read()
                buffer+=char
                if char=='M':   #mouse
                    b = self.ord(self.read(bin=True))-32
                    x = self.ord(self.read(bin=True))-32
                    y = self.ord(self.read(bin=True))-32
                    buffer+=f'\x1b[M{b};{x};{y}'
                    mouse=[b, x, y]
                else:           #other ansi sequence
                    while char>='0' and char<='9' or char==';':
                        char = self.read()
                        buffer+=char
        # Restore the original settings of the terminal
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, filedescriptors)
        fcntl.fcntl(sys.stdin.fileno(), fcntl.F_SETFL, flags & ~os.O_NONBLOCK)
        key=self.keymap.get(str(buffer))
        return key or str(buffer)

class widget():
    def __init__(self, x=1, y=1, w=1, h=1, fg=7, bg=0, key=None, action=None):
        self.fg0=7
        self.bg0=0
        self.invert=False
        self.key=key
        self.action=action
        self.minW=1
        self.minH=1
        self.t=termcontrol()
        self.kb=termKeyboard()
        self.setSize(x, y, w, h)
        self.setColors(fg, bg)
        self.widgetList=[]
        self.outstream=None #sys.stdout
        self.focus=None
        self.parent=None

    def __del__(self):
        pass

    def checkWidgetEvents(self, key, w):
        if key!='':
            esc='\x1b'
            print(f'{self.t.gotoxy(10,19)}  {key.replace(esc, "<")}         ')
        if w.key==key:
            if f'{type(self.action)}' in [ "function", "<class 'method'>" ]:
                self.invert=True
                self.action()
            else:
                print(f'{self.t.gotoxy(10,20)}invalid action for "{key}" type: {type(self.action)}             ')
        for cw in w.widgetList:
            cw.checkWidgetEvents(key, cw)

    def guiLoop(self):
        go=True
        self.t.disable_cursor()
        self.t.enable_mouse()
        buffercache=""
        while go:
            buffer=self.draw()
            if buffer != buffercache:
                buffercache=buffer
                print(buffer, end='')
            #print("\n"*3)
            key=self.kb.read_keyboard_input()
            self.checkWidgetEvents(key, self)
        self.t.enable_cursor()
        self.t.disable_mouse()

    def quit(self):
        exit(0)

    def setColors(self, fg, bg):
        self.fg, self.bg=fg, bg

    def setSize(self, x, y, w, h): #should always be okay
        if x<1:
            x=1
        if y<1:
            y=1
        scr=self.t.get_terminal_size()
        if w<self.minW:
            w=self.minW
        if h<self.minH:
            h=self.minH
        if x>scr['columns']-self.minW:
            x=scr['columns']-self.minW
        if y>scr['rows']-self.minH:
            y=scr['rows']-self.minH
        if w>scr['columns']-x+1:
            w=scr['columns']-x+1
        if h>scr['rows']-y+1:
            h=scr['rows']-y+1
        self.x=x
        self.y=y
        self.w=w
        self.h=h

    def addWidget(self, widget):
        widget.parent=self
        widget.fg0=self.fg
        widget.bg0=self.bg
        self.widgetList.append(widget)
        return self.widgetList[-1]

    def resize(self):
        for w in self.widgetList:
            w.resize()

    def drawChildren(self):
        buffer=''
        for w in self.widgetList:
            buffer+=w.draw()
        return buffer

    def draw(self):
        buffer=self.drawChildren()
        if self.outstream:
            self.outstream.write(buffer)
        return buffer

    def setFocus(self):
        pas

    def onFocus(self):
        pass

    def onDeFocus(self):
        pass

    def mouseEvent(self, x, y, buttons):
        pass

    def kbEvent(self, ch):
        pass

    def save(self, f):
        pass

    def load(self, f):
        pass

class widgetScreen(widget):
    def __init__(self, x, y, w, h, fg=7, bg=None, style=None):
        super().__init__(x=x, y=y, w=w, h=h, fg=fg, bg=bg)
        if theme:
            self.box=boxDraw(style=style, bgColor=self.bg, bg0=self.bg0)
        else:
            self.box=None
        self.style=style
        self.resize()
        self.feed=self.stream.feed

    def resize(self):
        super().resize()
        self.minW=5
        self.minH=5
        if self.box:
            self.screen = pyte.Screen(self.w-4, self.h-2)
            self.screen.screen_lines=self.h-2
        else:
            self.screen = pyte.Screen(self.w, self.h)
            self.screen.screen_lines=self.h
        self.screen.mode.add(pyte.modes.LNM)
        self.screen.encoding='utf-8'
        self.stream = pyte.Stream(self.screen)
        self.stream.write=self.stream.feed

    def draw(self):
        self.fg0=7
        self.bg0=0
        if self.parent:
            self.fg0=self.parent.fg
            self.bg0=self.parent.bg
        if self.style in [ 'outside' ]:
            self.bg0=self.bg
        if self.style in [ 'inside' ]:
            self.box.chars=self.box.chars[:4]+' '+self.box.chars[5:]
        buffer=''
        if(self.box):
            self.box.bg0=self.bg0
            buffer+=self.box.draw(self.x, self.y, self.w, self.h)
        self.stream.feed(self.t.ansicolor(self.fg, self.bg))
        self.stream.feed(super().drawChildren())
        self.stream.feed(self.t.gotoxy(1, self.screen.screen_lines))
        if self.box:
            buffer+=self.t.pyte_render(self.x+2, self.y+1, self.screen, fg=self.fg, bg=self.bg)
        else:
            buffer+=self.t.pyte_render(self.x, self.y, self.screen, fg=self.fg, bg=self.bg)
        if self.outstream:
            self.outstream.write(buffer)
        return buffer

    def input(self, str, maxlen=50):
        if maxlen < 1:
            maxlen=1
        if self.box:
            buffer = self.t.gotoxy(self.screen.cursor.x+self.x+2, self.screen.cursor.y+self.y+1)
        else:
            buffer = self.t.gotoxy(self.screen.cursor.x+self.x, self.screen.cursor.y+self.y)
        buffer+=str
        buffer +=self.t.reset()+' '*maxlen+self.t.left(maxlen)
        return input(buffer) #, maxlen=maxlen)

    def onFocus(self):
        pass

    def onDeFocus(self):
        pass

    def mouseEvent(self, x, y, buttons):
        pass

    def kbEvent(self, ch):
        pass

class widgetProgressBar(widget):
    def __init__(self, x, y, w, h, fg=7, bg=0, p0='\u2591', p1='\u2588', note=''):
        super().__init__(x=x, y=y, w=w, h=h, fg=fg, bg=bg)
        self.p0, self.p1=p0,p1
        self.note=note

    def draw(self, progress, total):
        buffer=self.t.gotoxy(self.x, self.y)
        buffer+=self.t.ansicolor(self.fg, self.bg)
        buffer+=self.note
        pct=0
        if total>0:
            pct=progress/total
        w=self.w-len(self.note)
        buffer +=self.p1*int((pct)*w)
        buffer +=self.p0*int(w-((pct)*w))
        buffer +=f'{self.t.gotoxy(self.x+len(self.note)+int(w/2)-5,self.y)}{progress}/{total}'
        return buffer

class widgetSlider(widget):
    def __init__(self, x, y, w, min=0, max=100, bg=233, barColor=238, labelColor=244, labelType="int", sliderColor=27, key=None, action=None):
        super().__init__(x=x, y=y, w=w, h=1, bg=bg, key=key, action=action)
        self.slider='\u2561\u2592\u255e'
        self.bar='\u2560\u2550\u2563'
        self.pos=min
        self.min=min
        self.max=max
        self.barColor=barColor
        self.labelColor=labelColor
        self.labelType=labelType
        self.sliderColor=sliderColor

    def draw(self):
        def hmsf(s):
            s = int(s)
            minutes = s / 60
            seconds = s % 60
            return f"{int(minutes):02d}:{int(seconds):02d}"
        if self.parent:
            self.fg0=self.parent.fg
            self.bg0=self.parent.bg
        LLabel=""
        RLabel=""
        if self.labelType=='int':
            LLabel=f'{self.min}'
            RLabel=f'{self.max}'
        elif self.labelType=='time':
            LLabel=f'{hmsf(self.min)}'
            RLabel=f'{hmsf(self.max)}' 
        else:
            pass
        barw=self.w-len(LLabel)-len(RLabel)-2-len(self.slider)-2
        pos=0
        if (self.max-self.min)>0:
            pos=self.pos/(self.max-self.min)
        spos=int((barw)*pos)
        buffer=self.t.gotoxy(self.x, self.y)
        buffer+=self.t.ansicolor(self.labelColor, self.bg0)
        buffer+=LLabel
        buffer+=self.t.ansicolor(self.barColor)
        buffer+=self.bar[0]
        buffer+=self.bar[1]*(spos)
        buffer+=self.t.ansicolor(self.sliderColor)
        buffer+=self.slider
        buffer+=self.t.ansicolor(self.barColor)
        buffer+=self.bar[1]*((barw-spos))
        buffer+=self.bar[2]
        buffer+=self.t.ansicolor(self.labelColor)
        buffer+=RLabel
        if self.outstream:
            self.outstream.write(buffer)
        return buffer

    def setValue(self, value):
        if value>=self.min and value<=self.max:
            self.pos=value
        if value>self.max:
            self.pos=self.max
        if value<self.min:
            self.pos=self.min

    def setMin(self, value):
        self.min=value
        self.setValue(self.pos)

    def setMax(self, value):
        self.max=value
        self.setValue(self.pos)

class widgetButton(widget):
    def __init__(self, x, y, w, h, fg=7, bg=None, style='curve', caption='Button', key=None, action=None, toggle=None):
        super().__init__(x=x, y=y, w=w, h=h, fg=fg, bg=bg, key=key, action=action)
        self.bg0=0
        self.fg0=7
        self.invert=False
        if theme:
            if style:
                self.box=boxDraw(style=style, bgColor=self.bg, bg0=self.bg0)
        else:
            self.box=None
        self.tint=None
        self.style=style
        self.caption=caption

    def draw(self):
        if self.parent:
            self.fg0=self.parent.fg
            self.bg0=self.parent.bg
        buffer=""
        if self.box:
            self.box.bg0=self.bg0
            self.box.tintFrame(self.tint)
            buffer+=self.box.draw(self.x, self.y, self.w, self.h, invert=self.invert)
        self.invert=False
        buffer+=self.t.gotoxy(self.x+self.w//2-(len(self.caption)//2), self.y+1)
        buffer+=self.t.ansicolor(self.fg, self.bg)
        buffer+=self.caption
        if self.key:
            buffer +=self.t.gotoxy(self.x+self.w//2-((len(self.key)+2)//2), self.y+2)
            buffer+=f'[{self.key}]'
        return buffer




from .Timer import Timer

from pydub import AudioSegment

STOP=0
PLAY=1
RECORD=2

class driver_audio():
    def __init__(self):
        self.endHandler=None
        self.status=STOP
        self.timer=Timer()
        self.cursor=0
        self.fps=44100
        self.channels=1
        self.sample_width=16//8
        self.record_buffer=None
        self.audio=None
        self.audio_start=0
        self.record_audio=None
        return None

    def load(self, filename):
        self.timer.clear()
        self.audio=AudioSegment.from_file(filename)
        self.setAudioProperties(fps=self.audio.frame_rate,
            channels=self.audio.channels, sample_width=self.audio.sample_width)
        self.audio_start=0
        return self.audio

    def play(self, start=0, end=0):
        self.status=PLAY
        #buffer=self.audio.get_array_of_samples()
        self.audio_start=self.get_cursor()
        self.timer.start(factor=self.audio.frame_rate, offset=-int(self.get_cursor()/self.audio.frame_rate))
        if end==0:
            pass
            #return sd.play(buffer[int(start):], self.audio.frame_rate)
        #return sd.play(buffer[int(start):int(end)], self.audio.frame_rate)

    def stop(self):
        #sd.stop()
        self.status=STOP
        if self.record_buffer is not None:
            frames=int(self.timer.get())
            self.record_audio=self.setAudio(self.record_buffer[:frames],
                self.fps,
                self.sample_width, self.channels)
            record_buffer=None
        self.timer.clear()
        self.audio_start=0

    def rec(self):
        self.status=RECORD
        self.audio_start=self.get_cursor()
        self.timer.start(factor=self.fps)
        len=self.fps*self.channels*60*10
        self.record_audio=None
        #self.record_buffer=sd.rec(len, self.fps, channels=self.channels)

    def wait(self):
        pass
        #return sd.wait()

    def save(self, filename):
        # Validate audio data
        if self.audio:
            self.audio.export(filename)

    def length_time(self):
        if self.audio:
            return self.length()/self.audio.frame_rate
        return 0

    def length(self):
        if self.audio:
            if self.status==RECORD:
                if self.timer.get():
                    return int(self.timer.get()+self.audio.frame_count())
            else:
                return int(self.audio.frame_count())
        else:
            if self.status==RECORD:
                return int(self.timer.get())
        return 0

    def get_cursor(self): #TODO FIXME find authoritative value
        if self.cursor>=self.length() and self.status==PLAY:
            if self.endHandler:
                self.endHandler()
        t=int(self.timer.get())
        if t>0:
            self.cursor=t
            if self.status==RECORD:
                self.cursor = t+self.audio_start
            if self.cursor>self.length():
                self.cursor=self.length()
        return self.cursor

    def get_cursor_time(self):
        fps=self.fps
        if self.audio:
            fps=self.audio.frame_rate
        if fps:
            return self.get_cursor()/fps #au.audio.frame_rate
        return 0

    def setAudio(self, buffer, fps, sample_width, channels):
        #return AudioSegment from buffer
        raise NotImplementedError(f"you must implement setAudio in {type(self)}")

    def setAudioProperties(self, fps=24000, channels=1, sample_width=16//8):
        self.fps=fps
        self.sample_width=sample_width
        self.channels=channels

    def concatenate(self, audiolist):
        full=None
        for a in audiolist:
            if full==None:
                full=a
            else:full=full+a
        return full

    def crop(self, start=None, end=None):
        fps=self.fps
        channels=self.channels
        sample_width=16//8
        if self.audio:
            fps=self.audio.frame_rate
            channels=self.audio.channels
            sample_width=self.audio.sample_width
        else:
            return None
        sf,ef=None, None
        if start is not None:
            sf=int(start)
        if end is not None:
            ef=int(end)
        if sf is not None and ef is not None:
            clip_frames=self.audio.get_array_of_samples()[sf:ef]
        elif sf is not None:
            clip_frames=self.audio.get_array_of_samples()[sf:]
        elif ef is not None:
            clip_frames=self.audio.get_array_of_samples()[:ef]
        else:
            clip_frames=self.audio.get_array_of_samples()
        #convert to AudioSegment
        audio_segment = AudioSegment(clip_frames.tobytes(), frame_rate=fps, 
            sample_width=sample_width, channels=channels)
        if audio_segment.frame_count()>0:
            return audio_segment
        return None




import sys,os,pydub,time, random
from optparse import OptionParser
np=None
try:
    from .driver_sounddevice_audio import driver_sounddevice_audio
    au=driver_sounddevice_audio()
    import numpy as np
    from scipy import signal as sp
except:
    print('missing sounddevice, numpy or scipy libraries.')
    from .driver_termux_audio import driver_termux_audio
    au=driver_termux_audio()
from pydub import AudioSegment

STOP=0
PLAY=1
RECORD=2

class pymms:
    def __init__(self):
        self.record_fps=24000
        self.record_channels=1
        self.record_sample_width=16//8
        self.selected=0
        self.selected_length=0
        self.stop()
        self.au=au

    def load(self, filename):
        self.stop()
        audio=au.load(filename)
        return {'title':filename, 'length':audio.duration_seconds, 
                'bitrate':audio.frame_rate, 'quality':audio.sample_width, 
                'channels':audio.channels}

    def save(self, filename):
        return au.save(filename)

    def playpause(self):
        if au.status in [ PLAY, RECORD ]:
            self.pause()
        elif au.status==STOP:
            self.play()

    def record(self):
        if au.status==STOP:
            au.setAudioProperties(fps=self.record_fps,
                channels=self.record_channels,
                sample_width=self.record_sample_width)
            c=au.cursor
            print(f'\nCursor={c/self.record_fps}')
            s=int(self.selected)
            sl=int(self.selected_length)
            if sl>0:
                self.pre=au.crop(None, s)
                self.post=au.crop(s+sl, None)
            else:
                self.pre=au.crop(None, c)
                self.post=au.crop(c, None) 
            if self.pre:
                print('\npre exists.')
            if self.post:
                print('post exists')
            au.rec()

    def play(self):
        if au.status==STOP and au.audio:
            c=self.get_cursor()
            s=int(self.selected)
            sl=int(self.selected_length)
            if sl==0:
                au.play(start=c)
            else:
                au.play(start=c, end=s+sl)

    def pause(self):
        if au.status==PLAY:
            au.stop()
        elif au.status==RECORD:
            length=int(au.timer.get())
            au.stop()
            if au.record_audio:
                #au.record_audio.get_array_of_samples()
                self.record_fps=au.record_audio.frame_rate
                self.record_sample_width=au.record_audio.sample_width
                self.record_channels=au.record_audio.channels
            else:
                print(f'\n\nMissing record audio')
                return
            print(f"Length:{length}")
            if self.pre:
                au.cursor=self.pre.frame_count()+au.record_audio.frame_count()
            else:
                au.cursor=au.record_audio.frame_count()
            self.selected=0
            self.selected_length=0
            ca=None
            if self.pre and self.post:
                print('\n\n\npre+rec+post')
                ca=au.concatenate((self.pre, au.record_audio, self.post))
            elif self.pre:
                ca=au.concatenate((self.pre, au.record_audio))
                print('\n\n\npre+rec     ')
            elif self.post:
                ca=au.concatenate((au.record_audio, self.post))
                print('\n\n\nrec+post    ')
            else:
                ca=au.record_audio
                print('\n\n\nrec         ')
            au.audio=ca
            self.pre, self.post = None, None

    def stop(self):
        self.pause()
        self.selected=0
        self.selected_length=0
        au.cursor=0

    def seek_time(self, time):
        if au.audio:
            self.seek(time*au.audio.frame_rate)

    def seek(self, frame):
        playing=au.status==PLAY
        if playing:
            self.pause()
        l=au.length()
        if au.status==STOP:
            if frame<0:
                frame=0
            if frame>l:
                time=l
            au.cursor=frame
        if au.cursor>l:
            au.cursor=l
        if playing:
            self.play()

    def seekFwd_time(self, time):
        if au.audio:
            self.seekFwd(time*au.audio.frame_rate)

    def seekFwd(self, frames):
        self.seek(au.cursor+frames)

    def seekBack_time(self, time):
        if au.audio:
            self.seekBack(time*au.audio.frame_rate)

    def seekBack(self, frames):
        self.seek(au.cursor-frames)

    def select_time(self, s, sl):
        self.select(s*au.audio.frame_rate, sl*au.audio.frame_rate)

    def select(self, s, sl):
        if au.status==STOP:
            if s<0:
                s=0
            if sl<0:
                sl=0
            if sl>0:
                l=au.length()
                if s<l:
                    s=l
                if s+sl>l:
                    sl=l-s
                self.selected=s
                self.selected_length=sl

    def clear(self):
        if self.selected_length==0:
            self.buffer=[]
        else:
            self.clear_selected()

    def clear_selected(self):
        self.seek(self.selected)
        s=int(self.selected)
        sl=int(self.selected_length)
        pre=self.buffer[:s]
        post=self.buffer[s+sl:]
        if len(pre)>0:
            if len(post)>0:
                self.buffer=au.concatenate((pre, post))
            else:
                self.buffer=pre
        else:
            if len(post)>0:
                self.buffer=post
            else:
                self.buffer=[]

    def get_cursor_time(self):
        return au.get_cursor_time()

    def get_cursor(self):
        return au.get_cursor()

    def length(self):
        return au.length()

    def length_time(self):
        return au.length_time()

    def denoise(self):
        au.noiseFilter()

    def normalize(self):
        pass

def main():
    return

if __name__ == "__main__":
    main()



#!/usr/bin/python

import sys,os,time, random
from optparse import OptionParser
from gm_termcontrol.termcontrol import termcontrol, pyteLogger, boxDraw, widget, widgetScreen
from gm_termcontrol.termcontrol import widgetProgressBar, widgetSlider, widgetButton
from gm_pymms.pymms import pymms

"""
         0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
U+250x   ─   ━   │   ┃   ┄   ┅   ┆   ┇   ┈   ┉   ┊   ┋   ┌   ┍   ┎   ┏
U+251x   ┐   ┑   ┒   ┓   └   ┕   ┖   ┗   ┘   ┙   ┚   ┛   ├   ┝   ┞   ┟
U+252x   ┠   ┡   ┢   ┣   ┤   ┥   ┦   ┧   ┨   ┩   ┪   ┫   ┬   ┭   ┮   ┯
U+253x   ┰   ┱   ┲   ┳   ┴   ┵   ┶   ┷   ┸   ┹   ┺   ┻   ┼   ┽   ┾   ┿
U+254x   ╀   ╁   ╂   ╃   ╄   ╅   ╆   ╇   ╈   ╉   ╊   ╋   ╌   ╍   ╎   ╏
U+255x   ═   ║   ╒   ╓   ╔   ╕   ╖   ╗   ╘   ╙   ╚   ╛   ╜   ╝   ╞   ╟
U+256x   ╠   ╡   ╢   ╣   ╤   ╥   ╦   ╧   ╨   ╩   ╪   ╫   ╬   ╭   ╮   ╯
U+257x   ╰   ╱   ╲   ╳   ╴   ╵   ╶   ╷   ╸   ╹   ╺   ╻   ╼   ╽   ╾   ╿
U+258x   ▀   ▁   ▂   ▃   ▄   ▅   ▆   ▇   █   ▉   ▊   ▋   ▌   ▍   ▎   ▏
U+259x   ▐   ░   ▒   ▓   ▔   ▕   ▖   ▗   ▘   ▙   ▚   ▛   ▜   ▝   ▞   ▟
U+25Ax   ■   □   ▢   ▣   ▤   ▥   ▦   ▧   ▨   ▩   ▪   ▫   ▬   ▭   ▮   ▯
U+25Bx   ▰   ▱   ▲   △   ▴   ▵   ▶   ▷   ▸   ▹   ►   ▻   ▼   ▽   ▾   ▿
U+25Cx   ◀   ◁   ◂   ◃   ◄   ◅   ◆   ◇   ◈   ◉   ◊   ○   ◌   ◍   ◎   ●
U+25Dx   ◐   ◑   ◒   ◓   ◔   ◕   ◖   ◗   ◘   ◙   ◚   ◛   ◜   ◝   ◞   ◟
U+25Ex   ◠   ◡   ◢   ◣   ◤   ◥   ◦   ◧   ◨   ◩   ◪   ◫   ◬   ◭   ◮   ◯
U+25Fx   ◰   ◱   ◲   ◳   ◴   ◵   ◶   ◷   ◸   ◹   ◺   ◻   ◼   ◽  ◾  ◿
"""

STOP=0
PLAY=1
RECORD=2

def minsec(s):
    s=int(s)
    mins=int(s/60)
    secs=int(s%60)
    return f'{mins:02d}:{secs:02d}'

class termplayer(widget):
    def __init__(self, x=1, y=1, w=80, h=15, mode='play', files=[], script=""): 
        self.icons={}
        self.icons['prev']     = {"label":'\u23ee',   "key":'[', 'action':self.prev}
        self.icons['prev']     = {"label":'\u25ae'+'\u25c0'*2, "key":'[', 'action':self.prev}
        self.icons['next']     = {"label":'\u23ed',   "key":']', 'action':self.next}
        self.icons['next']     = {"label":'\u25b6'*2+'\u25ae', "key":']', 'action':self.next}
        self.icons['play']     = {"label":'\u25b6',   "key":'P', 'action':self.play}
        self.icons['pause']    = {"label":'\u25ae'*2, "key":'p', 'action':self.pause}
        self.icons['play/pause']={"label":'\u25b6'+'\u25ae'*2, "key":'p', 'action':self.playpause}
        self.icons['stop']     = {"label":'\u25a0',   "key":'s', 'action':self.stop}
        self.icons['record']   = {"label":'\u25cf',   "key":'r', 'action':self.record}
        self.icons['eject']    = {"label":'\u23cf',   "key":'j', 'action':self.eject}
        self.icons['shuffle']  = {"label":'\u292e',   "key":'S', 'action':self.shuffle}
        self.icons['repeat']   = {"label":'\u21bb',   "key":'R', 'action':self.repeat}
        self.icons['seek']     = {"label":'',         "key":'k', 'action':self.seek}
        self.icons['seek-']    = {"label":'\u25c0'*2, "key":'-', 'action':self.seekBack}
        self.icons['seek+']    = {"label":'\u25b6'*2, "key":'+', 'action':self.seekFwd}
        self.icons['playlist'] = {"label":'\u2263',   "key":'L', 'action':self.togglePlayList}
        self.icons['denoise']  = {"label":'\u2593\u2592\u2591', "key":'N', 'action':self.denoise}
        self.icons['normalize']= {"label":'\u224b',   "key":'Z', 'action':self.normalize}
        self.icons['quit']=      {"label":'Quit',     "key":'q', 'action':self.quit}
        super().__init__(x=x, y=y, w=w, h=h)
        self.showPlayList=False
        self.clearPlayList=False
        self.playlist=files
        self.playlistinorder=files.copy()
        self.playListInfo={}
        self.repeat=False
        self.mode=mode
        for f in files:
            self.playListInfo[f]=self.mediaInfo(f)
        if self.mode=='record':
            if len(self.playlist)!=1:
                print("Record mode must reference one audio filename.")
                exit(1)
        else:
            pass
        self.filename=files[0]
        self.player=pymms()
        self.player.au.endHandler=self.endHandler
        if self.mode=='play':
            self.load(self.filename)
        self.script=script
        self.frame=0
        self.anim="\\-/|"
        self.x=x
        self.y=y
        self.w=w
        self.h=h
        self.playerbox=widgetScreen(self.x, self.y, self.w, self.h, bg=234, fg=15, style='outside')
        self.playlistbox=widgetScreen(self.x, self.y+self.h, self.w, self.h, bg=234, fg=15, style='outside')
        self.addWidget(self.playerbox)
        self.addWidget(self.playlistbox)
        boxHeight=7
        timeBoxW=30
        self.timeBox=widgetScreen(2, 1, timeBoxW, boxHeight, bg=233, fg=27, style='inside')
        self.playerbox.addWidget(self.timeBox)
        self.infoBox=widgetScreen(2+timeBoxW+2, 1, self.w-4-(timeBoxW+4), boxHeight, bg=233, fg=27, style='inside')
        self.playerbox.addWidget(self.infoBox)
        self.timeBox.box.tintFrame('#555')
        self.infoBox.box.tintFrame('#555')
        self.slider=widgetSlider(2, boxHeight+1, self.w-(2*2), 0, self.player.length(), labelType='time' , key='k')
        self.playerbox.addWidget(self.slider)
        self.addButtons(mode)

    def addButtons(self,mode):
        playbuttons=['prev', 'play/pause', 'stop', 'next', '', 'shuffle', 'repeat', 'playlist', '', 'quit']
        recordbuttons=['seek-', 'play/pause', 'stop', 'record', 'seek+', '', 'denoise', 'normalize', '', 'quit']
        buttons=playbuttons
        if mode=='record':
            buttons=recordbuttons
        else: 
            buttons=playbuttons
        self.btn={}
        btnX=2
        btnY=10
        btnW=7
        btnH=4 
        x=0
        for label in buttons:
            if self.icons.get(label):
                i=self.icons[label]
                toggle=None
                if label in ['shuffle', 'playlist', 'repeat']:
                    if label=='shuffle':
                        toggle=not (self.playlist==self.playlistinorder)
                    if label=='playlist':
                        toggle=self.showPlayList
                    if label=='repeat':
                        toggle=self.repeat
                self.btn[label]=widgetButton(x*btnW+btnX, btnY, btnW, btnH, fg=27, bg=233, caption=i['label'], key=i['key'], action=i['action'], toggle=toggle)
                self.playerbox.addWidget(self.btn[label])
            x+=1

    def mediaInfo(self, f):
        title=f
        length=0
        bitrate=0
        quality=0
        channels=1
        info={'title':title, 'length':length, 'bitrate':bitrate, 'quality':quality, 'channels':channels}
        self.playListInfo[f]=info
        return info

    def drawBigString(self, s):
        chars={}
        chars['Resolution']='5x4'
        chars['0']= " ▄▄  "\
                    "█  █ "\
                    "▄  ▄ "\
                    "▀▄▄▀ "
        chars['1']= "     "\
                    "   █ "\
                    "   ▄ "\
                    "   ▀ "
        chars['2']= " ▄▄  "\
                    "   █ "\
                    "▄▀▀  "\
                    "▀▄▄  "
        chars['3']= " ▄▄  "\
                    "   █ "\
                    " ▀▀▄ "\
                    " ▄▄▀ "
        chars['4']= "     "\
                    "█  █ "\
                    " ▀▀▄ "\
                    "   ▀ "
        chars['5']= " ▄▄  "\
                    "█    "\
                    " ▀▀▄ "\
                    " ▄▄▀ "
        chars['6']= " ▄▄  "\
                    "█    "\
                    "▄▀▀▄ "\
                    "▀▄▄▀ "
        chars['7']= " ▄▄  "\
                    "   █ "\
                    "   ▄ "\
                    "   ▀ "
        chars['8']= " ▄▄  "\
                    "█  █ "\
                    "▄▀▀▄ "\
                    "▀▄▄▀ "
        chars['9']= " ▄▄  "\
                    "█  █ "\
                    " ▀▀▄ "\
                    " ▄▄▀ "
        chars[':']= "     "\
                    "  ●  "\
                    "  ●  "\
                    "     "
        chars[' ']= "     "\
                    "     "\
                    "     "\
                    "     "
        col,row=chars['Resolution'].split('x')
        col=int(col)
        row=int(row)
        buffer=""
        for y in range(row):
            for c in s:
                fc=chars.get(c)
                if fc:
                    buffer+=fc[y*col:(y+1)*col]
                else:
                    buffer+=" "
            buffer+='\n'
        return buffer

    def drawMultiLine(self, x, y, s):
        lines=s.split('\n')
        dy=0
        buffer=""
        for l in lines:
            buffer+=self.t.gotoxy(x, y+dy)
            buffer+=l
            dy=dy+1
        return buffer

    def draw(self):
        t=self.player.get_cursor_time()
        self.slider.setValue(t)
        self.slider.setMax(max(self.player.length_time(), t))
        timestr=self.drawBigString(minsec(t))
        buffer=''
        fg=27
        if self.player.au.status==RECORD:
            fg=196
        else:
            fg=27
        if self.mode=='record':
            self.infoBox.feed(self.t.clear())
            self.infoBox.feed(self.t.gotoxy(1, 1))
            self.infoBox.feed(self.t.ansicolor(27))
            self.infoBox.feed(self.script)
            pass
        elif self.mode=='play':
            self.infoBox.feed(self.t.clear())
            self.infoBox.feed(self.t.ansicolor(27))
            title=""
            title=f'{self.playlist.index(self.filename)+1}. '
            if self.playListInfo[self.filename]: 
                title+=self.playListInfo[self.filename]['title']
                title+=f' ({minsec(self.playListInfo[self.filename]["length"])})'
            self.infoBox.feed(self.t.gotoxy(1, 1))
            self.infoBox.feed(title)
            quality=0
            bitrate=0
            channels=0
            if self.playListInfo[self.filename]: 
                quality=int(self.playListInfo[self.filename]["quality"])
                bitrate=int(self.playListInfo[self.filename]["bitrate"])
                channels=int(self.playListInfo[self.filename]["channels"])
            self.infoBox.feed(self.t.gotoxy(1, 3))
            self.infoBox.feed(f'{int(quality)}kbps')
            self.infoBox.feed(self.t.gotoxy(14, 3))
            self.infoBox.feed(f'{bitrate}kHz')
            self.infoBox.feed(self.t.gotoxy(31, 3))
            if channels==0:
                self.infoBox.feed(f'No Audio')
            elif channels==1:
                self.infoBox.feed(f'    Mono')
            elif channels==2:
                self.infoBox.feed(f'  Stereo')
            else:
                self.infoBox.feed(f' Stereo+')
        self.timeBox.feed(self.t.ansicolor(fg,233, bold=True))
        self.timeBox.feed(self.t.clear())
        self.timeBox.feed(self.drawMultiLine(30-(5*5)-2, 1, timestr))
        self.timeBox.feed(self.t.gotoxy(1, 3))
        if self.player.au.status==PLAY:
            i=self.icons['play']
            self.timeBox.feed(self.t.ansicolor(46, 233, bold=True))
            self.timeBox.feed(i['label'])
        elif self.player.au.status==RECORD:
            i=self.icons['record']
            self.timeBox.feed(self.t.ansicolor(196, 233, bold=True))
            self.timeBox.feed(i['label'])
        elif self.player.au.status==STOP:
            i=self.icons['stop']
            self.timeBox.feed(self.t.ansicolor(27, 233, bold=True))
            self.timeBox.feed(i['label'])
        rcolor=234
        scolor=234
        if self.repeat:
            rcolor=27
        if self.playlist!=self.playlistinorder:
            scolor=27
        i=self.icons['repeat']
        self.timeBox.feed(self.t.gotoxy(1, 4))
        self.timeBox.feed(self.t.ansicolor(rcolor, 233, bold=True))
        self.timeBox.feed(i['label'])
        i=self.icons['shuffle']
        self.timeBox.feed(self.t.gotoxy(1, 5))
        self.timeBox.feed(self.t.ansicolor(scolor, 233, bold=True))
        self.timeBox.feed(i['label'])
        buffer+=self.playerbox.draw()
        if self.showPlayList:
            self.playlistbox.feed(f'{self.t.clear()}')
            startline=self.playlist.index(self.filename)
            if startline+self.h-3>len(self.playlist):
                startline=len(self.playlist)-(self.h-2)
            if startline<0:
                startline=0
            for n in range(self.h-2):
                color=27
                if n+startline<len(self.playlist):
                    f=self.playlist[startline+n]
                    if f==self.filename:
                        color=46
                    self.playlistbox.feed(f'{self.t.gotoxy(1,n+1)}')
                    self.playlistbox.feed(f'{self.t.ansicolor(color)}')
                    title=f
                    tm="00:00"
                    if self.playListInfo[f]:
                        title=f"{self.playListInfo[f]['title']}"
                        tm=f"{minsec(self.playListInfo[f]['length'])}"
                    self.playlistbox.feed(f'{n+startline+1}. {title}')
                    self.playlistbox.feed(f'{self.t.gotoxy(self.w-3-len(tm),n+1)}')
                    self.playlistbox.feed(f'{tm}')
                else:
                    self.playlistbox.feed(f'{self.t.gotoxy(1,n+1)} ') 
            buffer+=self.playlistbox.draw()
        else:
            if self.clearPlayList:
                self.clearPlayList=False
                buffer+=self.t.reset()
                for y in range(self.y+self.h,self.y+2*self.h):
                    buffer+=self.t.gotoxy(self.x, y)
                    for x in range(0, self.w):
                        buffer+=' '
        #print(self.anim[self.frame % len(self.anim)])
        self.frame +=1
        return buffer

    def togglePlayList(self):
        self.showPlayList=not self.showPlayList
        if not self.showPlayList:
            self.clearPlayList=True

    def load(self, filename):
        info=self.player.load(filename)
        self.playListInfo[filename]=info
        self.filename=filename

    def save(self, filename):
        self.player.save(filename)

    def quit(self):
        if self.mode=='record':
            #TODO save prompt
            if(1):
                self.save(self.filename)
        self.stop()
        quit()

    def next(self):
        i=self.playlist.index(self.filename)
        i+=1
        if i>=len(self.playlist):
            i=0
        self.load(self.playlist[i])
        if self.mode=='play':
            self.play()

    def prev(self):
        i=self.playlist.index(self.filename)
        i-=1
        if i<0:
            i=len(self.playlist)-1
        self.load(self.playlist[i])
        if self.mode=='play':
            self.play()

    def endHandler(self):
        if self.player.au.status==PLAY:
            if self.mode=='play':
                if self.repeat:
                    self.stop()
                    self.play()
                else:
                    self.next()
            else:
                self.player.pause()

    def shuffle(self):
        if self.playlist==self.playlistinorder:
            random.shuffle(self.playlist)
        else:
            self.playlist=self.playlistinorder.copy()

    def repeat(self):
        self.repeat = not self.repeat

    def seek(self, pos=0):
        self.player.seek_time(pos)

    def seekFwd(self):
        self.player.seekFwd_time(10)

    def seekBack(self):
        self.player.seekBack_time(10)

    def eject(self):
        pass

    def play(self):
        self.player.play()

    def pause(self):
        self.player.pause()

    def playpause(self):
        self.player.playpause()

    def stop(self):
        self.player.stop()

    def record(self):
        self.player.record()

    def denoise(self):
        self.player.denoise()

    def normalize(self):
        self.player.normalize()

def main():
    parser=OptionParser(usage="usage: %prog [options] AUDIO_FILES")
    parser.add_option("-r", "--record", action='store_true', dest="record", 
            default=False, help="Record mode.")
    parser.add_option("-v", "--verbose", dest="debug", default="info",
            help="Show debug messages.[debug, info, warning]")
    parser.add_option("-s", dest="script", default="No script was given.",
            help="Script for record mode.")
    parser.add_option("-x", dest="x", default=1, help="Top left coordinate.")
    parser.add_option("-y", dest="y", default=1, help="Top left coordinate.")

    (options, args)=parser.parse_args()
    if len(args)==0:
        parser.print_help()
        return
    mode='play'
    if options.record:
        mode='record'
    tp=termplayer(x=int(options.x), y=int(options.y), mode=mode, script=options.script, files=args)
    tp.guiLoop()
    return

if __name__ == "__main__":
    main()



import sys,os,shutil,pydub
from pydub import AudioSegment

termux_play=shutil.which('termux-media-player')
termux_record=shutil.which('termux-microphone-record')
termux_api=os.environ.get('TERMUX_API_VERSION')
record_temp_file='._rec.wav'
play_temp_file='._play.wav'
err=False
if not termux_api:
    err=True
    print('termux api is not installed.')
else:
    import termux
if not termux_play:
    err=True
    print('Missing termux-media-player')
if not termux_record:
    err=True
    print('Missing termux-microphone-record')
if not err:
    print('termux OK')

 
from pydub import AudioSegment

from .driver_audio import driver_audio

class driver_termux_audio(driver_audio):
    def __init__(self):
        super().__init__()
        self.lastaction=''

    def play_file(self, fn):
        termux.Media.play(fn)

    def rec_file(self, fn, fps=24000):
        self.fps=fps
        termux.Microphone.record(fn, rate=fps)

    def play(self, start=0, end=0):
        super().play(start=start, end=end)
        buffer=self.audio.get_array_of_samples()
        if self.lastaction=='':
            if end==0:
                buf=buffer[int(start):]
            else:
                buf=buffer[int(start):int(end)]
            audio_segment = pydub.AudioSegment(buf,
                frame_rate=self.audio.frame_rate, 
                sample_width=self.audio.sample_width, 
                channels=self.audio.channels)
            audio_segment.export(play_temp_file)
            self.lastaction='play'
            self.play_file(play_temp_file)

    def stop(self):
        if self.lastaction=='play':
            self.lastaction=''
            termux.Media.control("stop")
            os.remove(play_temp_file)
        elif self.lastaction=='record':
            self.lastaction=''
            termux.Microphone.stop()
            self.record_audio=AudioSegment.from_file(record_temp_file)
            self.record_buffer=self.record_audio.get_array_of_samples()
            os.remove(record_temp_file)
        super().stop()

    def rec(self):
        super().rec()
        if self.lastaction=='':
            self.rec_file(record_temp_file, fps=self.fps)
            self.lastaction='record'
        #return self.record_buffer

    def wait(self):
        super().wait()
        return sd.wait()

    def setAudio(self, buffer, fps, sample_width, channels):
        # Create a pydub AudioSegment 
        return AudioSegment(buffer.tobytes(),
            frame_rate=fps, sample_width=sample_width, channels=channels)




#!/usr/bin/python3
from setuptools import setup, find_packages
import shutil

with open('README.md') as f:
    long_description = f.read()
shutil.copyfile('gm_pymms/termplayer.py', 'gm_pymms/pymms')

setup(
    name='gm_pymms',
    version='0.0.2',
    license='GPL3',
    url='https://github.com/gretchycat/pymms',
    author='Gretchen Maculo',
    author_email='gretchen.maculo@gmail.com',
    description='python xmms inspired media player/recorder',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        'pydub',
        'gm_termcontrol'
    ],
    tests_require=[
    ],
    scripts=['gm_pymms/pymms']
)
